package akhlys.fastdirsync;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;

import javax.swing.JOptionPane;

/**
 * Compares the source directory with the target one
 * @author akhlys
 *
 */
public class DifferenceFinderThread extends NicelyCancelableThread<List<FileDifference>, Void> {
	FastDirSync parent = null;
	Path sourceDir = null;
	Path targetDir = null;
	
	/**
	 * new instance
	 * @param parent
	 * @param source
	 * @param target
	 */
	public DifferenceFinderThread(FastDirSync parent, Path source, Path target) {
		this.parent = parent;
		this.sourceDir = source;
		this.targetDir = target;
	}
	
	@Override
	protected List<FileDifference> doInBackground() throws Exception {
		this.parent.progressBar.setString("Comparing..");
		// toRealPath so subfiles aren't "C:file" but "C:\file" if the path is C: (surprisingly both work just fine in code, but looks weird in output).
		// Also normalizeds the path (C:\bla\..\dir -> C:\dir) and checks if the path actually exists
		List<FileDifference> differences = searchDifferences(this.sourceDir.toRealPath(), this.targetDir.toRealPath());
		if (differences == null) return null; // aborted
		
		return differences;
	}
	
	private List<FileDifference> searchDifferences(Path sourceDir, Path targetDir) {
		List<FileDifference> differences = new ArrayList<>();
		
		// at least in one instance, "system volume information" reported as "can read" but listfiles still returned null
		List<Path> sourceFiles = listFiles(sourceDir);
		List<Path> unaccountedForTargetFiles = listFiles(targetDir); // at the start all will be unaccounted for
		if (sourceFiles == null) {
			System.out.println("Error, cannot read from source directory, skip: " + sourceDir.toString());
			return differences;
		}
		if (unaccountedForTargetFiles == null) {
			System.out.println("Error, cannot read from target directory, skip: " + targetDir.toString());
			return differences;
		}

		for (Path fileToSearchFor : sourceFiles) {
			if (this.isAborted()) return null;
			
			Path evaluatedTargetFile = targetDir.resolve(fileToSearchFor.getFileName().toString()).toAbsolutePath();
			
			if (unaccountedForTargetFiles.contains(evaluatedTargetFile)) { // same as Files.exist() at this moment
				// remove from targetFiles, so that in the end the only files remaining are those to be deleted
				unaccountedForTargetFiles.remove(evaluatedTargetFile);
				
				if (Files.isSymbolicLink(fileToSearchFor) || Files.isSymbolicLink(evaluatedTargetFile)) continue;
				
				if ((Files.isRegularFile(fileToSearchFor) && Files.isDirectory(evaluatedTargetFile))) {
					// name exists, but wrong type, meaning the one on the target has to be removed and added anew
					differences.add(new FileDifference(fileToSearchFor, evaluatedTargetFile, FileDiffType.CHANGED_TYPE_F2D));
					continue;
				}
				if ((Files.isDirectory(fileToSearchFor) && Files.isRegularFile(evaluatedTargetFile))) {
					// name exists, but wrong type, meaning the one on the target has to be removed and added anew
					differences.add(new FileDifference(fileToSearchFor, evaluatedTargetFile, FileDiffType.CHANGED_TYPE_D2F));
					continue;
				}
				
				if (Files.isDirectory(fileToSearchFor)) {
					// path exists on both source and target, and is a directory on both
					// recursively call funct for files in dir
					List<FileDifference> childDifferences = searchDifferences(fileToSearchFor, evaluatedTargetFile);
					if (childDifferences == null) return null; // was aborted
					differences.addAll(childDifferences);
				} else {
					
					try {
						if (Files.size(fileToSearchFor) != Files.size(evaluatedTargetFile)) {
							// file exists but is different in size
							differences.add(new FileDifference(fileToSearchFor, evaluatedTargetFile, FileDiffType.CHANGED));
						}
					} catch (IOException e) {
						System.err.println("Could not check size of " + fileToSearchFor.toString() + ", " + e.getMessage());
					}
				}
			} else {
				// file or directory is missing
				if (Files.isSymbolicLink(fileToSearchFor)) continue;
				differences.add(new FileDifference(fileToSearchFor, evaluatedTargetFile, FileDiffType.NEW));
			}
		}
		
		// only files remaining in targetFiles are not in sourceFiles
		for (Path f : unaccountedForTargetFiles) differences.add(new FileDifference(null, f, FileDiffType.DELETED));
		
		return differences;
	}
	
	private List<Path> listFiles(Path dir) {
		if (!Files.isReadable(dir)) return null;
		List<Path> list = new ArrayList<>();
		try {
			Files.list(dir).forEach(entry -> list.add(entry.toAbsolutePath()));
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
		return list;
	}
	
	
	@Override
	protected void done() {
		// reset gui
		this.parent.progressBar.setIndeterminate(false);
		this.parent.lockGUIForTask(false);
		this.parent.currentTask = null;
		
		if (this.isAborted()) {
			System.out.println("aborted");
			return;
		}
		
		// fill table
		List<FileDifference> tableData = null;
		try {
			tableData = get();
		} catch (InterruptedException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this.parent, e.getLocalizedMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			return;
		} catch (ExecutionException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this.parent, e.getLocalizedMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		this.parent.sourceOfLastCompare = this.sourceDir;
		this.parent.targetOfLastCompare = this.targetDir;

		if (tableData.size() == 0) {
			((FileDifferencesTableModel)this.parent.resultTable.getModel()).setDifferences(null, null, null);
			this.parent.syncButton.setEnabled(false);
			JOptionPane.showMessageDialog(this.parent, "No differences found", "Directories equal", JOptionPane.INFORMATION_MESSAGE);
		} else {
			((FileDifferencesTableModel)this.parent.resultTable.getModel()).setDifferences(tableData, sourceDir, targetDir);
		}
	}
}