package akhlys.fastdirsync;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.stream.Stream;

import javax.swing.JOptionPane;

/**
 * Compares the source directory with the target one
 * @author akhlys
 *
 */
public class SynchronizationThread extends NicelyCancelableThread<Boolean, Void> {
	FastDirSync parent = null;
	Path sourceDir = null;
	Path targetDir = null;
	List<FileDifference> differences = null;
	
	/**
	 * new instance
	 * @param parent
	 * @param source
	 * @param target
	 * @param differences 
	 */
	public SynchronizationThread(FastDirSync parent, Path source, Path target, List<FileDifference> differences) {
		this.parent = parent;
		this.sourceDir = source.normalize(); // normalizing should not be necessary, but eh, just in case
		this.targetDir = target.normalize();
		this.differences = differences;
	}
	
	@Override
	protected Boolean doInBackground() {
		this.parent.progressBar.setIndeterminate(false);
		double numOfChanges = this.differences.size();
		double numOfCurrentChange = 1;
		
		boolean retVal = true;
		
		for (FileDifference difference : this.differences) {
			if (this.isAborted()) return null;
			this.parent.progressBar.setValue((int)((numOfCurrentChange/numOfChanges)*100));
			
			// do what has to be done
			switch(difference.diffType) {
				case DELETED:
					this.parent.progressBar.setString("Deleting " + difference.targetFile.getFileName().toString());
					System.out.println("Deleting " + difference.targetFile.getFileName().toString());
					// dirs are never writeable and not really needed, del error would be catched anyway
					/*if (!Files.isWritable(targetFile)) {
						retVal = false;
						break;
					}*/
					try {
						deleteFileOrDir(difference.targetFile);
					} catch (IOException e) {
						e.printStackTrace();
						retVal = false;
					}
					break;
				case CHANGED:
				case CHANGED_TYPE_D2F:
				case CHANGED_TYPE_F2D:
					this.parent.progressBar.setString("Replacing " + difference.sourceFile.getFileName().toString());
					System.out.println("Replacing " + difference.sourceFile.getFileName().toString());
					if (!Files.isWritable(difference.targetFile) || !Files.isReadable(difference.sourceFile)) {
						retVal = false;
						break;
					}
					try {
						deleteFileOrDir(difference.targetFile);
					} catch (IOException e) {
						e.printStackTrace();
						retVal = false;
						break;
					}
					// not a fall trough to new because of the progress bar
					try {
						if (Files.isDirectory(difference.sourceFile)) {
							System.out.println("Copying dir " + difference.sourceFile.getFileName().toString());
							copyDirectory(difference.sourceFile, difference.targetFile);
						} else {
							System.out.println("Copying file " + difference.sourceFile.getFileName().toString());
							Files.copy(difference.sourceFile, difference.targetFile);
						}
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
						retVal = false;
					}
					break;
				case NEW:
					this.parent.progressBar.setString("Copying " + difference.sourceFile.getFileName().toString());
					try {
						if (Files.isDirectory(difference.sourceFile)) {
							System.out.println("Copying dir " + difference.sourceFile.getFileName().toString());
							copyDirectory(difference.sourceFile, difference.targetFile);
						} else {
							System.out.println("Copying file " + difference.sourceFile.getFileName().toString());
							Files.copy(difference.sourceFile, difference.targetFile);
						}
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
						retVal = false;
					}
					break;
			}
			++numOfCurrentChange;
		}
		return retVal;
	}
	
	public static void deleteFileOrDir(Path file)  throws IOException {
		if (!Files.exists(file)) {
			System.out.println("File is already deleted: " + file.toString());
			return;
		}
		if (Files.isDirectory(file)) {
			try (Stream<Path> walk = Files.walk(file)) {
			    walk.sorted(Comparator.reverseOrder())
			        .map(Path::toFile)
			        .peek(System.out::println)
			        .forEach(File::delete);
			}
		} else {
			Files.delete(file);
		}
	}
	
	/**
     * Copies a directory to a new location.
     * 
     * @param srcDir  the validated source directory, must not be <code>null</code>
     * @param destDir  the validated destination directory, must not be <code>null</code>
     * @throws IOException if an error occurs
     */
    private void copyDirectory(Path srcDir, Path destDir) throws IOException {
        // recurse
    	ArrayList<Path> files = new ArrayList<>();
        Files.list(srcDir).forEach(entry -> files.add(entry));
        if (Files.exists(destDir)) {
            if (Files.isDirectory(destDir) == false) {
                throw new IOException("Destination '" + destDir + "' exists but is not a directory");
            }
        } else {
            Files.createDirectories(destDir);
        }
        if (Files.isWritable(destDir) == false) {
            throw new IOException("Destination '" + destDir + "' cannot be written to");
        }
        for (Path file : files) {
        	if (this.isAborted()) return;
            Path copiedFile = destDir.resolve(file.getFileName().toString());
            if (Files.isDirectory(file)) {
                copyDirectory(file, copiedFile);
            } else {
            	Files.copy(file, copiedFile);
            }
        }

        // Do this last, as the above has probably affected directory metadata
        Files.setLastModifiedTime(destDir, Files.getLastModifiedTime(srcDir));
    }
	
	@Override
	protected void done() {
		// reset gui
		this.parent.progressBar.setIndeterminate(false);
		this.parent.progressBar.setValue(0);
		this.parent.lockGUIForTask(false);
		this.parent.currentTask = null;
		((FileDifferencesTableModel)this.parent.resultTable.getModel()).setDifferences(null, null, null);
		
		if (this.isAborted()) {
			System.out.println("aborted");
			return;
		}
		
		boolean returnValue = true;
		try {
			returnValue = get();
		} catch (InterruptedException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this.parent, e.getLocalizedMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			return;
		} catch (ExecutionException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this.parent, e.getLocalizedMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			return;
		}

		if (returnValue) {
			JOptionPane.showMessageDialog(this.parent, "Synchronized!", "Finished", JOptionPane.INFORMATION_MESSAGE);
		} else {
			JOptionPane.showMessageDialog(this.parent, "Could not fix all differences due to access errors. Compare anew for those files.", "Finished with problems", JOptionPane.INFORMATION_MESSAGE);
		}
	}
}